# THIS FILE IS GENERATED FROM ORANGE-WIDGET-BASE-ML SETUP.PY
short_version = '4.17.0'
version = '4.17.0'
full_version = '4.17.0'
git_revision = 'ba995ee38dc40baaddd0a248c8dc05c4f108ca18'
release = True

if not release:
    version = full_version
    short_version += ".dev"
